require 'hutzbot/version'
require 'httparty'

module Hutzbot
  class APIError < StandardError; end

  class Response
    attr_accessor :id
    attr_accessor :speaker
    attr_accessor :speaker_name
    attr_accessor :statement
    attr_accessor :metadata

    def initialize attrs
      @id, @speaker, @speaker_name, @statement, @metadata = *attrs.values_at('id', 'speaker', 'speaker_name', 'statement', 'metadata')
    end

    def system?
      speaker=='system'
    end
  end

  class ValidResponse
    attr_accessor :answer
    attr_accessor :response

    def initialize attrs
      @answer, @response = *attrs.values_at('answer', 'response')
    end
  end

  class Conversation
    attr_accessor :id
    attr_accessor :created_at
    attr_accessor :current_state
    attr_accessor :taxonomy
    attr_accessor :daily_counter
    attr_accessor :tagged_document
    attr_accessor :user
    attr_accessor :application
    attr_accessor :responses
    attr_accessor :valid_responses

    def initialize attrs
      @id, @current_state, @taxonomy, @tagged_document, @user, @application, @daily_counter = *attrs.values_at('id', 'current_state', 'taxonomy', 'tagged_document', 'user', 'application', 'daily_counter')
      @created_at = Date.parse(attrs['created_at'])
      @responses = attrs['responses'].map{ |r| Response.new(r) } if attrs['responses']
      @valid_responses = attrs['valid_responses'].map{ |vr| ValidResponse.new(vr) } if attrs['valid_responses'] 
    end

    def closed?
      current_state=='closed'
    end
  end

  class Taxonomy
    attr_accessor :id
    attr_accessor :name

    def initialize attrs
      @id, @name = *attrs.values_at('id', 'name')
    end
  end

  class TaggedDocument
    attr_accessor :id
    attr_accessor :title

    def initialize attrs
      @id, @title = *attrs.values_at('id', 'title')
    end
  end


  class Hutzbot
    def initialize logger, host, user, key, application
      @logger = logger
      @host = host
      @user = user
      @key = key
      @application = application
      @options = { headers: { 'X-User-Email' => @user, 'X-User-Token' => @key, 'Accept' => 'application/json' } }
    end

    def logger
      @logger
    end

    def start_conversation
      taxonomy = get_default_taxonomy
      tagged_document = get_default_tagged_document
      conversation = create_conversation({ taxonomy_id: taxonomy.id, tagged_document_id: tagged_document.id, application: @application })
    end

    def get_conversations metadata=nil
      request = [@host, 'conversations'].join('/')
      request += "?metadata=#{metadata}" unless metadata.nil?
      response = HTTParty.get(request, @options)
      check_errors response
      JSON.parse(response.body).map{ |c| Conversation.new(c) }
    end

    def get_conversation id
      response = HTTParty.get("#{@host}/conversations/#{id}", @options)
      check_errors response
      Conversation.new(JSON.parse(response.body))
    end 

    def tag_responses conversation_id, metadata
      response = HTTParty.patch("#{@host}/conversations/#{conversation_id}/tag_responses", @options.merge({ query: { metadata: metadata } }))
      check_errors response
      Conversation.new(JSON.parse(response.body))
    end

    def add_response conversation_id, answer
      response = HTTParty.patch("#{@host}/conversations/#{conversation_id}/add_response", @options.merge({ query: { answer: answer } }))
      check_errors response
      Conversation.new(JSON.parse(response.body))
    end

    def get_default_taxonomy
      response = HTTParty.get("#{@host}/taxonomies/default", @options)
      check_errors response
      Taxonomy.new(JSON.parse(response.body))
    end

    def get_default_tagged_document
      response = HTTParty.get("#{@host}/tagged_documents/default", @options)
      check_errors response
      TaggedDocument.new(JSON.parse(response.body))
    end

    def create_conversation params
      response = HTTParty.post("#{@host}/conversations", @options.merge({ query: { conversation: params.merge({ application: @application }) } }))
      check_errors response
      Conversation.new(JSON.parse(response.body))
    end

    def converse metadata, body
      if metadata.nil? || metadata.empty?      
        # start a conversation via the API
        logger.info 'Starting a new conversation'
        conversation = start_conversation
        logger.info "Conversation ID: #{conversation.id}"
      else
        logger.info "Finding a conversation by metadata (#{metadata})"

        # find the conversation by metadata
        conversations = get_conversations metadata
        unless conversations.empty?
          # found a matching conversation
          conversation = conversations.first
          logger.info "Found conversation (#{conversation.id})"

          # now get the full conversation (more attributes than available in index call)
          conversation = get_conversation conversation.id

          # get the answer index
          matches = /(-?\d+)/.match(body)
          if matches.nil?
            # invalid answer
            logger.warn("No answer found in body (#{body})")
          else
            answer_index = matches[1].to_i
            logger.info("Found an answer in the message with index #{answer_index} in (#{conversation.valid_responses})")

            matching_answer = conversation.valid_responses[answer_index-1]
            logger.info("That answer is #{matching_answer.answer} (#{matching_answer.response})")
            conversation = add_response conversation.id, matching_answer.answer
          end
        end
      end

      if conversation
        logger.info "Preparing response"
        responses = conversation.responses.select{ |r| r.system? && (r.metadata.nil? || r.metadata.empty?) }
        contents = responses.map do |r|
          "<p>#{r.statement}</p>"  
        end.join("<br>")
        logger.debug "Response is: #{contents}"

        logger.info "Preparing prompt"
        prompt = nil
        unless conversation.closed?
          prompt = conversation.valid_responses.map do |valid_response|
            "<li>#{valid_response.response}</li>"
          end.join
          prompt = "<p>Please make your selection: </p><ol>#{prompt}</ol>"
        end
        logger.debug "Prompt is: #{prompt}"

        return conversation, contents, prompt
      else
        return nil, ""
      end
    end

    private
    def check_errors response
      case response.code
      when 200..299
        "Success"
      when 400..499
        raise APIError, "HutzbotAPI Error: #{response.code}"
      else
        raise APIError, "HutzbotAPI Error: #{response.code}"
      end
    end
  end
end